/* This is free software. See ../COPYING for details. */
#include <stdlib.h>
#include <util/delay.h>
#include <uWireM/uWireM.h>
#include "hmc5883l.h"

#define HMC5883L_I2C_ADDR 0x1e

/* Number of samples averaged per reading: */
#define CFG_AVG_1 0x00
#define CFG_AVG_2 0x01
#define CFG_AVG_4 0x02
#define CFG_AVG_8 0x03

/* Load bias direction: */
#define CFG_BIAS_NONE 0x00
#define CFG_BIAS_POS  0x01
#define CFG_BIAS_NEG  0x02

/* Gain (in LSb per Gauss): */
#define CFG_GAIN_1370 0x00
#define CFG_GAIN_1090 0x01
#define CFG_GAIN_820  0x02
#define CFG_GAIN_660  0x03
#define CFG_GAIN_440  0x04
#define CFG_GAIN_390  0x05
#define CFG_GAIN_330  0x06
#define CFG_GAIN_230  0x07

/* Configuration choices: */
#define CFG_AVG CFG_AVG_8
#define CFG_BIAS CFG_BIAS_NONE
#define CFG_GAIN CFG_GAIN_1090
#define GAIN (1.0/1090)

/* Register addresses on HMC5883L: */
#define REG_CRA 0  /* config A */
#define REG_CRB 1  /* config B */
#define REG_MR 2   /* mode */
#define REG_DATA 3 /* registers 3-8 inclusive are data, read as a block */
#define REG_SR 9   /* status */
#define REG_IRA 10 /* ID A = 'H' */
#define REG_IRB 11 /* ID B = '4' */
#define REG_IRC 12 /* ID C = '3' */

static uint8_t _buf[7];

/*** hmc5883l_read() -- read sensor values from HMC5883L

Read the sensor values from the HMC5883L.

Note: Results are undefined if HMC5883L has not yet been configured
using hmc5883l_config().

Arguments:
   p -- pointer to data structure into which sensor data is to be
      stored, or NULL if the readings are to be discarded
***/
void hmc5883l_read(hmc5883l_pos_t * const p) {

   /* Start a measurement in single measurement mode: */
   _buf[0] = uWireM_addr_send(HMC5883L_I2C_ADDR);
   _buf[1] = REG_MR;
   _buf[2] = 0x01;
   uWireM_xfer(_buf, 3);

   /* Tell the device we want to start reading at DXRA (register 3): */
   _buf[1] = REG_DATA;
   uWireM_xfer(_buf, 2);

   /* We could poll the status register here, but to save some space,
      we'll just block for a minimum of (1Hz/s)/160Hz = 6250us, after
      which we know the reading will be ready. */
   _delay_us(6250);

   /* Read all six positions: */
   _buf[0] = uWireM_addr_recv(HMC5883L_I2C_ADDR);
   uWireM_xfer(_buf, 7);

   if(!p) return;
   p->x = _buf[1]; p->x = (p->x << 8) | _buf[2];
   p->y = _buf[3]; p->y = (p->y << 8) | _buf[4];
   p->z = _buf[5]; p->z = (p->z << 8) | _buf[6];
}

/*** hmc5883l_init() -- configure HMC5883L sensor

Performs initialization and configuration on HMC5883L sensor.
***/
void hmc5883l_init(void) {
   /* Set config register A for desired averaging and bias. We leave
      the output rate all-bits-zero since we're just going to be using
      single measurement mode: */
   _buf[0] = uWireM_addr_send(HMC5883L_I2C_ADDR);
   _buf[1] = REG_CRA;
   _buf[2] = CFG_AVG << 5 | CFG_BIAS;
   uWireM_xfer(_buf, 3);

   /* Set config register B for desired gain: */
   _buf[1] = REG_CRB;
   _buf[2] = CFG_GAIN << 5;
   uWireM_xfer(_buf, 3);

   /* Set mode register to force idle mode: */
   _buf[1] = REG_MR;
   _buf[2] = 0x03; /* IDLE */
   uWireM_xfer(_buf, 3);

   // FIXME: could the above three writes be combined?

   /* If the gain was changed, the next reading will use the old gain.
      Perform a reading and discard the results so the caller doesn't
      have to worry about this. */
   hmc5883l_read(NULL);
}
